<?php

//	Here is a bit of test code. If this file is called as the "action" for an HTML form
//	that accepts a typed character string in a field named "str", this will display the
//	encoded string and the result of decoding it. You can change the character to be
//	used as the escape character in the toZvvyCode function call.

$str = $_POST['str'];
echo ('Original string: '.$str."<p>");
$code = toZvvyCode($str, 0x2f);
echo ('Encoded string: '.$code."<p>");

$decod = fromZvvyCode($code);
echo ('Decoded: '.$decod['text']."<p>");
echo ('Error: '.$decod['error']."<p>");

exit;



//
// ZvvyCode:  Universal safe text encoding method
//
// 		By David R. Woolley
// 		Thinkofit, Inc.
//		www.thinkofit.com
//
//
// ZvvyCode is a text encoding method that converts virtually any text into a format that can
// be handled by virtually any programming language without the danger of there being characters
// embedded in the encoded text that have syntactic meaning in the language. An encoded string
// consists only of the ASCII letters (a-z and A-Z) and digits (0-9), plus one special escape
// character which is designated at the time of encoding and may be anything other than the
// ASCII letters and digits. An encoded string can be decoded without advance knowledge of the
// escape character, because the escape character is read from the beginning of the encoded string.
//
// The original ASCII letters (a-z and A-Z) and digits (0-9) are kept in their original form.
// ALL other characters, including spaces, punctuation, extended ASCII codes, Unicode characters,
// etc. are converted to their hexadecimal character codes bounded on each end by a uniform
// escape character. The escape character may be any character except the ASCII letters and digits,
// so you can choose an escape character that will not cause trouble in your application.
//
// An encoded string begins with an 8-character sequence consisting of two instances of the escape
// character used in encoding the string, the letters "zvvy", and two more instances of the escape
// character. The decoder reads these characters to determine (1) whether the string is, in fact,
// a ZvvyCode string, and (2) what character to interpret as the escape character for the decoding
// process.
//
// If the first eight characters do not follow this pattern exactly, or if the characters in the
// positions designated for the escape characters are ASCII letters or digits, then the string
// is considered to be NOT ZvvyCode, and is returned as-is, without decoding.
//
// Example: Suppose you choose a slash as the escape character to encode this text:
//
// 		I said, "Joe's bag is 3/4 full!"
//
// The ZvvyCode encoding would be:
//
//		//zvvy//I/20/said/2c//20//22/Joe/27/s/20/bag/20/is/20/3/2f/4/20/full/21//22/
//
//	Note that it doesn not matter that the original text has a slash in it, even though the slash
//	is to be used as the escape character. The slash is simply converted to /2f/.
//
//	The escape character does not have to be a visible character at all. It can be, for example,
//	0 (ASCII null), 127 (ASCII del), 27 (ASCII esc), 255 (the last extended ASCII character)
//	or even a large Unicode character value. The only restriction is that it cannot be one of
//	the ASCII upper or lower case letters or numeric digits.
//





//
//	toZvvyCode: Convert a string to ZvvyCode encoding.
//		written by David R. Woolley, Thinkofit, Inc.	www.thinkofit.com
//
//
//	Arguments:
//		str		String to be encoded
//		escChr	Character code to use as the escape character, given as a numeric value.
//				(E.g., 92 or 0x5c to use backslash as the escape.)
//				MUST NOT BE A LETTER (a-z, A-Z) OR DIGIT (0-9). Any other character code is valid.
//
//	Returns:
//		ZvvyCode encoded string, or empty string if original string is empty
//
//

function toZvvyCode ($str, $escChr) {

	if ($str == '') {
		// If string is empty, return empty.
		return ('');
	}

	$lth = strlen($str);
	$escChrStr = chr($escChr);

	// Begin with two escape characters, the "zvvy" identifier, and two more escapes.
	$rs = $escChrStr.$escChrStr."zvvy".$escChrStr.$escChrStr;

	for ($i = 0; $i < $lth; $i++) {
		$ch = ord($str{$i});
		if (($ch >= 0x30 && $ch <= 0x39) || ($ch >= 0x41 && $ch <= 0x5A) || ($ch >= 0x61 && $ch <= 0x7A)) {
			// Copy a-z, A-Z, and 0-9 as-is.
			$rs .= $str{$i};
		}
		else {
			// Output a hex character value surrounded by escape characters.
			$rs .= $escChrStr . dechex($ch) . $escChrStr;
		}
	}
	return ($rs);
}


//
//	fromZvvyCode: Convert a ZvvyCode encoded string back to its original value.
//
//	Arguments:
//		str		String to be decoded
//
//	Returns:
//		Associative array with two elements:
//			text:	Decoded string
//			error:	-1 if successful
//					 0 if the string was determined to be not encoded and is returned unchanged
//					 2 or greater if the encoded string is malformed
//
//	An encoded string is considered malformed if any of the following occur:
//		* A value enclosed between two escape characters is not a valid hexadecimal number
//		* An opening escape character is not matched by a closing escape - i.e., the string ends
//		  before a terminating escape character is found.
//
//	In case of a malformed string, the returned array will contain:
//			text:	As much of the string as could be decoded before the error was encountered
//			error:	Position within the string where a problematic escape sequence began
//

function fromZvvyCode ($str) {

	$lth = strlen($str);
	$retobj = array('text' => '', 'error' => -1);

	if ($lth < 8) {
		// If string is less than eight characters, it cannot be ZvvyCode, so return as is.
		$retobj['text'] = $str;
		$retobj['error'] = 0;
		return $retobj;
	}

	$escChrStr = $str{0};	// First character should be the escape character used to encode this string
	$escChr = ord($escChrStr);
	if (($escChr >= 0x30 && $escChr <= 0x39) || ($escChr >= 0x41 && $escChr <= 0x5A) || ($escChr >= 0x61 && $escChr <= 0x7A)) {
		// First character is a letter or digit, not a valid escape character -- means this is not ZvvyCode so return as is.
		$retobj['text'] = $str;
		$retobj['error'] = 0;
		return $retobj;
	}
	if ($escChrStr != $str{1} || $escChrStr != $str{6} || $escChrStr != $str{7} || substr($str,2,4) != "zvvy") {
		// First 8 characters do not follow ZvvyCode pattern -- means string is not ZvvyCode, so return as is.
		$retobj['text'] = $str;
		$retobj['error'] = 0;
		return $retobj;
	}

	echo ('Esc char:'.$escChrStr.'<p>');

	$i = 8;			// Start decoding after the first two characters
	$retstr = '';	// Initialize return string
	while ($i < $lth) {
		$escPos = strpos($str, $escChrStr, $i);
		if ($escPos === false) {
			// No more instances of the escape string found, so append remainder of string and we are done.
			$retobj['text'] = $retstr.substr($str, $i);
			return $retobj;
		}
		$retstr = $retstr.substr($str, $i, $escPos - $i);	// Append all regular characters up to where escape was found
		// echo ('i='.$i.' escPos='.$escPos.'<br>');
		$i = $escPos + 1;						// Point to first character beyond the escape
		$escPos = strpos($str, $escChrStr, $i);	// Find terminating escape
		// echo ('&nbsp;&nbsp;i='.$i.' escPos='.$escPos);
		if ($escPos) {
			$escVal = '0x'.substr ($str, $i, $escPos - $i);
			if (!is_numeric($escVal)) {
				// Value in the escape sequence is not a valid number
				$retobj['text'] = $retstr;
				$retobj['error'] = $i - 1;
				return $retobj;
			}
			$escVal += 0;	// PHP seems to need this to actually turn the hex value into a number.
			$retstr .= chr($escVal);
			// echo (' escVal='.$escVal.' chr='.chr($escVal).' retstr='.$retstr.'<br>');
			$i = $escPos + 1;
		}
		else {

			// Escape sequence was started but string ended with no terminating escape found.
			// This indicates a malformed string.
			// Return decoded text up to the beginning of the unterminated escape sequence,
			// and set error value to the character position where the escape sequence began.

			$retobj['error'] = $i - 1;
			$retobj['text'] = $retstr;
			return $retobj;
		}
	}
	$retobj['text'] = $retstr;
	$retobj['error'] = -1;
	return $retobj;
}

?>